#!/usr/bin/env python3

import os, sys, re, stat, pathlib, argparse, random, types, shutil
import threading, subprocess, tempfile, io, base64, zlib, webbrowser
import socketserver, http.server, urllib.parse, html, time, unicodedata

wx_version = 0
try:
    import wx
    wx_version = int(wx.__version__.split(".")[0])
except:
    sys.stderr.write("the wx package has to be installed\n")
    sys.exit(1)
if wx_version < 4:
    sys.stderr.write("wxPython has to be at major version 4\n")
    sys.exit(1)

SERVER_DOMAIN = "localhost"
PORT_MIN = 1024
PORT_MAX = 65535
SERVER_SETTING_PORT = 12001
SERVER_SETTING_TIMEOUT = 2
SERVER_SERVING_PORT = 12000
SERVER_SERVING_PORT_MIN = 15000
SERVER_SERVING_PORT_MAX = 40000
LOGGING_WEB_DEFAULT = False
LOGGING_SET_DEFAULT = False
HOME_PATH_RE = re.compile(re.escape(str(pathlib.Path.home())+"/"))
MAPPING_STATIC = "~/.webmixer/mapping_static"
MAPPING_STATIC_WIN = "~/webmixer_mapping.ini"
MAPPING_SESSION = "~/.webmixer/mapping_session"
MAPPING_SESSION_WIN = "~/webmixer_session.ini"
APPLICATION_NAME = "Webmixer"
APPLICATION_DESC = "personal webserver"
APPLICATION_LINK = "https://webmixer.tangloid.net"
APPLICATION_LICENSE = "GPL 2+"
APPLICATION_VERSION = "0.21"
RANDOM_STRING_LENGTH = 16
HEXDIGITS = "0123456789abcdef"
ALLOWED_DYNAMIC_URLS = re.compile("\\/[" + HEXDIGITS + "]{8,32}\\/$")
ICONDATA = """
eJztk1lP4zAUhZ+nv8IsBVIITZzEToEiEpu+zWgeR0KlKqXse8u+/Hbs44tLGWBGGomnsRDK
53t8l2O3XmO/fn5ntXplMOwO93ust9e9YDU27A+GnSji22nn5ux4o82a7L4ymXBm/kSDxZML
lUn2rcd+nJ707fei+Z7KCqnKdYvzFtNEpppbXLOYlDLRiE4h2pJpWVqcBgqZqMJilcRcK4sz
JBYKqWrAhsyUtNiEWJnMEIdAbRCZlyGOTRvC4ipwXSYlCi0AC5lrjDOLsy1zFpkDnwqZJwi5
zi0+kTjVSHXv28CAbWSWkqvU4ibMEbIoWxbrJI5dV3PUZKxQqAMszYCJxSWITc+u7gp5lTnx
BjIrqZ2Tj9RVrHH2wZ9FNPaInrlFIfPMXVmCaGEGBKaUiuvMYoZCDVmWqCvQRiS0QlTSgGmh
LeZUiNpo0IDcPYaIbqF0tndJzJ3tWz4aWeyRk0LBum1/R3Cj723HgDtoQxqfgbt0v9J5tUc3
GGtcyj6JlfP5gJ5ZXKLJQ2AqhevqiApxZ92xd7Jh8YTcaCngKTJHMnPP7My/HETPfc/w6oIm
ilyTA6rLFUYY+ltAz5eImrPuIV35rnD2mt4GVzDnxv+OkOoW0cwUwrx3JI4dFh5RqHSZhXRO
KjevyB1qiHNRFKjL/rC+RLM4rplfm/pNMl2dea2pNcNweXVMsTAbhBOvNU+hWfftkWKzbk6F
wdxI0wmxgqUVp9h4fHA7sdfwJKQVpJnZbK+9cCheNDIP/GazEXVHVN0a9dPb9tthf/Q5uzPm
z+4o/0vdvf23Hh4cHo1Jjk/e8/k0RCNn9t/5xUd3MViiMsNL9pGGsStorsccf6u5geb2LzR3
/zX/oKkGdhWfakpll/5U8976Us3jcuUZlF+Axw==
"""

def is_win_form():
    if "win" in sys.platform:
        if "darwin" not in sys.platform:
            return True
    return False

def is_linux_like():
    os_desc = wx.GetOsDescription().lower()
    return True if (("linux" in os_desc) or ("bsd" in os_desc)) else False

def is_hidden(path, subpath, allow_dot=False, check_tilde=False):
    if check_tilde and path.endswith("~"):
        return True
    if not os.access(path, os.R_OK):
        return True
    stat_res = os.stat(path)
    if hasattr(stat_res, "st_file_attributes") and hasattr(stat, "FILE_ATTRIBUTE_HIDDEN"):
        if bool(stat_res.st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN):
            return True
    if hasattr(stat_res, "st_flags") and hasattr(stat, "UF_HIDDEN"):
        if bool(stat_res.st_flags & stat.UF_HIDDEN):
            return True
    if subpath.startswith(".") and (not allow_dot):
        return True
    return False

def trim_subpath(subpath):
    return subpath.strip("/")

def is_dotted_inner(path_end):
    for path_part in path_end.split("/"):
        if path_part.startswith(".") and (path_part != "."):
            return True
    return False

def shorten_home_dir(path):
    match = HOME_PATH_RE.match(path)
    if match is None:
        return path
    return "~/" + path[match.end():]

class AppFrame(wx.Frame):
    def __init__(self, properties):
        super().__init__(parent=None, title=(APPLICATION_NAME + ": " + APPLICATION_DESC))
        self.app_properties = properties
        self.web_port_str = str(self.app_properties["port_serving"])
        self.base_serving_url = "http://" + SERVER_DOMAIN + ":" + self.web_port_str

        self.ui = types.SimpleNamespace()
        ui = self.ui

        ui.panel = wx.Panel(self)
        ui.panel_sizer = wx.BoxSizer(wx.VERTICAL)

        ui.top_sizer = wx.BoxSizer(wx.HORIZONTAL)
        ui.panel_sizer.Add(ui.top_sizer, 0, wx.ALL | wx.EXPAND, 0)

        if self.app_properties["with_static_map"]:
            label_static_mapping_txt = "static mapping:"
        else:
            label_static_mapping_txt = "no static mapping"
        ui.label_static_mapping = wx.StaticText(ui.panel, label=label_static_mapping_txt, size=(-1, -1))
        ui.top_sizer.Add(ui.label_static_mapping, 0, wx.ALL | wx.EXPAND, 5)

        ui.top_spacer = wx.StaticText(ui.panel, label=" ", size=(-1, -1))
        ui.top_sizer.Add(ui.top_spacer, 1, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)

        ui.label_web_port = wx.StaticText(ui.panel, label="web serving at port " + self.web_port_str, size=(-1, -1))
        ui.top_sizer.Add(ui.label_web_port, 0, wx.ALL | wx.EXPAND, 5)

        ui.app_link_button = wx.BitmapButton(ui.panel, bitmap=wx.ArtProvider.GetBitmap(wx.ART_TIP), size=(32,-1))
        ui.top_sizer.Add(ui.app_link_button, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 1)
        ui.app_link_button.Bind(wx.EVT_BUTTON, self.on_app_link_click)
        ui.app_link_button.SetToolTip("Open web site of the Webmixer application")

        ui.close_button = wx.BitmapButton(ui.panel, bitmap=wx.ArtProvider.GetBitmap(wx.ART_QUIT), size=(32,-1))
        ui.top_sizer.Add(ui.close_button, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 0)
        ui.close_button.Bind(wx.EVT_BUTTON, self.on_button_close)
        ui.close_button.SetToolTip("Quit the Webmixer application")

        ui.list_static = None
        if self.app_properties["with_static_map"]:
            ui.list_static = wx.ListView(ui.panel, size=(-1, 120), style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
            ui.list_static.Bind(wx.EVT_KILL_FOCUS, lambda ev: self.on_losing_focus(ui.list_static))
            ui.list_static.Bind(wx.EVT_SET_FOCUS, lambda ev: self.on_gaining_focus(ui.list_static))
            ui.list_static.Bind(wx.EVT_RIGHT_DOWN, lambda ev: None)
            ui.list_static.Bind(wx.EVT_RIGHT_UP, lambda ev: None)
            ui.list_static.Bind(wx.EVT_RIGHT_DCLICK, lambda ev: None)
            ui.list_static.SetToolTip("Ctrl+B copies the whole URL of selected mapping")

            ui.list_static.InsertColumn(0, "URL   (Ctrl+C,B)", width=150)
            ui.list_static.InsertColumn(1, "directory   (Ctrl+Shift+C)", width=300)
            ui.panel_sizer.Add(ui.list_static, 3, wx.ALL | wx.EXPAND | wx.CENTER, 5)
 
            map_index = 0
            for item in self.app_properties["mapping"]:
                ui.list_static.InsertItem(map_index, item["url"])
                ui.list_static.SetItem(map_index, 1, shorten_home_dir(item["dir"]))
                map_index += 1

        ui.middle_sizer = None
        ui.label_dynamic_mapping = None
        ui.app_link_label = None
        ui.restore_btn = None
        ui.clear_btn = None
        ui.list_dynamic = None
        if self.app_properties["with_dynamic_map"]:
            ui.middle_sizer = wx.BoxSizer(wx.HORIZONTAL)
            ui.panel_sizer.Add(ui.middle_sizer, 0, wx.ALL | wx.EXPAND, 0)

            ui.label_dynamic_mapping = wx.StaticText(ui.panel, label="dynamic mapping:", size=(-1, -1))
            ui.middle_sizer.Add(ui.label_dynamic_mapping, 2, wx.ALL | wx.EXPAND, 5)

            ui.app_link_label = wx.StaticText(ui.panel, label=" mapping set at port " + str(SERVER_SETTING_PORT), size=(-1, -1))
            ui.middle_sizer.Add(ui.app_link_label, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 5)

            ui.restore_btn = wx.BitmapButton(ui.panel, bitmap=wx.ArtProvider.GetBitmap(wx.ART_REDO), size=(32,-1))
            ui.middle_sizer.Add(ui.restore_btn, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 1)
            ui.restore_btn.Bind(wx.EVT_BUTTON, self.on_button_restore)
            if (
                (len(self.app_properties["session"]) > 0)
                or (len(self.app_properties["session_restore"]) == 0)
            ):
                ui.restore_btn.Disable()
            ui.restore_btn.SetToolTip("Restore the dynamic mapping")
            ui.clear_btn = wx.BitmapButton(ui.panel, bitmap=wx.ArtProvider.GetBitmap(wx.ART_CLOSE), size=(32,-1))
            ui.middle_sizer.Add(ui.clear_btn, 0, wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL, 0)
            ui.clear_btn.Bind(wx.EVT_BUTTON, self.on_button_clear)
            if len(self.app_properties["session"]) == 0:
                ui.clear_btn.Disable()
            ui.clear_btn.SetToolTip("Clear the dynamic mapping")

            ui.list_dynamic = wx.ListView(ui.panel, size=(-1, 150), style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
            ui.list_dynamic.Bind(wx.EVT_KILL_FOCUS, lambda ev: self.on_losing_focus(ui.list_dynamic))
            ui.list_dynamic.Bind(wx.EVT_SET_FOCUS, lambda ev: self.on_gaining_focus(ui.list_dynamic))
            ui.list_dynamic.Bind(wx.EVT_RIGHT_DOWN, lambda ev: None)
            ui.list_dynamic.Bind(wx.EVT_RIGHT_UP, lambda ev: None)
            ui.list_dynamic.Bind(wx.EVT_RIGHT_DCLICK, lambda ev: None)
            ui.list_dynamic.SetToolTip("Ctrl+B copies the whole URL of selected mapping")

            ui.list_dynamic.InsertColumn(0, "URL   (Ctrl+C,B)", width=150)
            ui.list_dynamic.InsertColumn(1, "directory   (Ctrl+Shift+C)", width=300)
            ui.panel_sizer.Add(ui.list_dynamic, 4, wx.ALL | wx.EXPAND | wx.CENTER, 5)

            with self.app_properties["maplock"]:
                self.app_properties["setting_envelope"].wx_frame = self
                self.app_properties["serving_envelope"].wx_frame = self
                map_index = 0
                for item in self.app_properties["session"]:
                    ui.list_dynamic.InsertItem(map_index, item["url"])
                    ui.list_dynamic.SetItem(map_index, 1, shorten_home_dir(item["dir"]))
                    map_index += 1

        key_ctrl_c_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_c, id=key_ctrl_c_id)
        key_ctrl_b_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_b, id=key_ctrl_b_id)
        key_ctrl_shift_c_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_shift_c, id=key_ctrl_shift_c_id)
        key_ctrl_r_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_r, id=key_ctrl_r_id)
        key_ctrl_k_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_k, id=key_ctrl_k_id)
        key_ctrl_q_id = wx.NewIdRef()
        self.Bind(wx.EVT_MENU, self.on_ctrl_q, id=key_ctrl_q_id)

        self.ui.accel_tbl = wx.AcceleratorTable([
            (wx.ACCEL_CTRL, ord('C'), key_ctrl_c_id),
            (wx.ACCEL_CTRL, ord('B'), key_ctrl_b_id),
            (wx.ACCEL_SHIFT|wx.ACCEL_CTRL, ord('C'), key_ctrl_shift_c_id),
            (wx.ACCEL_CTRL, ord('R'), key_ctrl_r_id),
            (wx.ACCEL_CTRL, ord('K'), key_ctrl_k_id),
            (wx.ACCEL_CTRL, ord('Q'), key_ctrl_q_id),
        ])
        self.SetAcceleratorTable(self.ui.accel_tbl)

        ui.panel.SetSizerAndFit(ui.panel_sizer)
        ui.sizer = wx.BoxSizer(wx.VERTICAL)
        ui.sizer.Add(ui.panel, 1, wx.ALL | wx.EXPAND | wx.CENTER, 5)
        self.SetSizerAndFit(ui.sizer)

        try:
            image_data = zlib.decompress(base64.b64decode(ICONDATA))
            image_stream = io.BytesIO(bytes(image_data))
            image = wx.Image(image_stream, wx.BITMAP_TYPE_ANY)
            bitmap = wx.Bitmap(image)
            icon = wx.Icon()
            icon.CopyFromBitmap(bitmap)
            self.SetIcon(icon)
        except:
            pass

        self.Bind(wx.EVT_CLOSE, self.on_close)
        screen_width, screen_height = wx.DisplaySize()
        cca_app_size = 480
        app_pos_horiz_default = 100
        app_pos_vert_default = 50
        app_pos_horiz = app_pos_horiz_default if (screen_width >= (cca_app_size + 2*app_pos_horiz_default)) else 0
        app_pos_vert = app_pos_vert_default if (screen_height >= (cca_app_size + 2*app_pos_vert_default)) else 0
        self.SetPosition(wx.Point(app_pos_horiz, app_pos_vert))
        self.Show()

    def update_session_list(self):
        if not self.app_properties["with_dynamic_map"]:
            return
        ui = self.ui
        with self.app_properties["maplock"]:
            map_count = len(self.app_properties["session"])
            if map_count > 0:
                ui.restore_btn.Disable()
                ui.clear_btn.Enable()
            else:
                self.ui.clear_btn.Disable()
            map_index = ui.list_dynamic.GetItemCount()
            if (map_index == 0) and (map_count > 0):
                self.app_properties["session_restore"] = []
            if map_index == map_count:
                return
            for item in self.app_properties["session"][map_index:]:
                self.app_properties["session_restore"].append(item)
                ui.list_dynamic.InsertItem(map_index, item["url"])
                ui.list_dynamic.SetItem(map_index, 1, shorten_home_dir(item["dir"]))
                map_index += 1
            if self.app_properties["session_path"] is not None:
                try:
                    with open(self.app_properties["session_path"], "w", encoding="utf-8") as session_fh:
                        for item in self.app_properties["session_restore"]:
                            session_fh.write(item["url"] + "\n")
                            session_fh.write(item["dir"] + "\n\n")
                except:
                    sys.stderr.write("cannot save current session into session file\n")
                    sys.stderr.write(self.app_properties["session_path"] + "\n")

    def on_app_link_click(self, ev):
        webbrowser.open(APPLICATION_LINK)

    def on_gaining_focus(self, widget):
        if widget in (self.ui.list_static, self.ui.list_dynamic):
            widget.Select(widget.GetFocusedItem(), True)

    def on_losing_focus(self, widget):
        if widget in (self.ui.list_static, self.ui.list_dynamic):
            widget.Select(widget.GetFirstSelected(), False)

    def get_focused_widget(self):
        return wx.Window.FindFocus()

    def on_ctrl_c(self, ev):
        widget = self.get_focused_widget()
        if widget in (self.ui.list_static, self.ui.list_dynamic):
            self.copy_list_item(widget, 0, False)

    def on_ctrl_b(self, ev):
        widget = self.get_focused_widget()
        if widget in (self.ui.list_static, self.ui.list_dynamic):
            self.copy_list_item(widget, 0, True)

    def on_ctrl_shift_c(self, ev):
        widget = self.get_focused_widget()
        if widget in (self.ui.list_static, self.ui.list_dynamic):
            self.copy_list_item(widget, 1)

    def on_ctrl_r(self, ev):
        if self.app_properties["with_dynamic_map"]:
            self.restore_session()

    def on_ctrl_k(self, ev):
        if self.app_properties["with_dynamic_map"]:
            self.clear_session()

    def on_ctrl_q(self, ev):
        self.close_app()

    def copy_via_xclip(self, text):
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(text.encode("utf-8"))
            tmpfile.flush()
            try:
                subprocess.run([self.app_properties["xclip_path"], tmpfile.name])
            except:
                pass

    def put_to_clipboard(self, text_to_copy, with_xclip):
        if text_to_copy and wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(text_to_copy))
            wx.TheClipboard.Flush()
            wx.TheClipboard.Close()
            if with_xclip and self.app_properties["use_xclip"]:
                self.copy_via_xclip(text_to_copy)

    def copy_list_item(self, widget, column, prepend_server=False):
        text_to_copy = ""
        index = widget.GetFirstSelected()
        if -1 < index:
            text_to_copy = widget.GetItemText(index, column)
            if prepend_server:
                middle_slash = "" if text_to_copy.startswith("/") else "/"
                text_to_copy = self.base_serving_url + middle_slash + text_to_copy
        self.put_to_clipboard(text_to_copy, True)

    def on_button_close(self, ev):
        self.close_app()

    def on_close(self, ev):
        self.close_app()

    def close_app(self):
        if self.app_properties["setting_envelope"] is not None:
            try:
                self.app_properties["setting_envelope"].server.shutdown()
            except:
                pass
        if self.app_properties["serving_envelope"] is not None:
            try:
                self.app_properties["serving_envelope"].server.shutdown()
            except:
                pass
        self.Destroy()

    def on_button_restore(self, ev):
        if not self.app_properties["with_dynamic_map"]:
            return
        self.restore_session()
        self.ui.panel.SetFocus()

    def restore_session(self):
        if not self.app_properties["with_dynamic_map"]:
            return
        with self.app_properties["maplock"]:
            ui = self.ui
            if len(self.app_properties["session_restore"]) == 0:
                return
            if len(self.app_properties["session"]) > 0:
                return
            ui.restore_btn.Disable()
            ui.clear_btn.Enable()
            map_index = 0
            for item in self.app_properties["session_restore"]:
                self.app_properties["session"].append(item)
                ui.list_dynamic.InsertItem(map_index, item["url"])
                ui.list_dynamic.SetItem(map_index, 1, shorten_home_dir(item["dir"]))
                map_index += 1

    def on_button_clear(self, ev):
        self.clear_session()
        self.ui.panel.SetFocus()

    def clear_session(self):
        if not self.app_properties["with_dynamic_map"]:
            return
        with self.app_properties["maplock"]:
            ui = self.ui
            if len(self.app_properties["session_restore"]) > 0:
                self.ui.restore_btn.Enable()
            ui.clear_btn.Disable()
            ui.list_dynamic.DeleteAllItems()
            empty_session = []
            self.app_properties["session"] = empty_session
            self.app_properties["setting_envelope"].session = empty_session
            self.app_properties["serving_envelope"].session = empty_session

class SettingRH(socketserver.BaseRequestHandler):
    def generate_random_string(self):
        return "".join([random.choice(HEXDIGITS) for ind in range(RANDOM_STRING_LENGTH)])

    def check_can_descend(self, dir_check, dir_end):
        if is_hidden(dir_check, trim_subpath(dir_end), True, False):
            return False
        if is_dotted_inner(dir_end):
            return False
        return True

    def update_setting(self, dir_check):
        envelope = self.server.envelope
        if not dir_check.endswith("/"):
            dir_check += "/"
        dir_check_re = re.compile(re.escape(dir_check))
        present_link = None
        generated_item = None
        with envelope.maplock:
            for mapset in [envelope.session, envelope.mapping]:
                for rule in mapset:
                    match = rule["dir_re"].match(dir_check)
                    if match is not None:
                        dir_end = dir_check[match.end():]
                        if not trim_subpath(dir_end):
                            present_link = rule["url"]
                            break
                        if self.check_can_descend(dir_check, dir_end):
                            if (not rule["url"].endswith("/")) and (not dir_end.startswith("/")):
                                dir_end = "/" + dir_end
                            elif (rule["url"].endswith("/")) and (dir_end.startswith("/")):
                                dir_end = dir_end.lstrip("/")
                            present_link = rule["url"] + dir_end
                            break
                if present_link is not None:
                    break
            if present_link is not None:
                return {"url": present_link, "is_new": False}

            while True:
                test_url_dir = "/" + self.generate_random_string() + "/"
                test_url_dir_re = re.compile(re.escape(test_url_dir))
                test_passed = True
                for item in envelope.session:
                    if item["url_re"].match(test_url_dir) or test_url_dir_re.match(item["url"]):
                        test_passed = False
                        break
                if test_passed:
                    for item in envelope.mapping:
                        if test_url_dir_re.match(item["url"]):
                            test_passed = False
                            break
                if test_passed:
                    generated_item = {
                        "url": test_url_dir,
                        "url_re": test_url_dir_re,
                        "dir": dir_check,
                        "dir_re": dir_check_re,
                    }
                    envelope.session.append(generated_item)
                    break
        return {"url": generated_item["url"], "is_new": True}

    def check_challenge(self, challenge, challenge_answer):
        answer_lines = challenge_answer.splitlines()
        if len(answer_lines) < 2:
            return False
        if answer_lines[0].upper().strip() != "MAP/1.0 ANSWER":
            return False
        answer = answer_lines[1].strip()
        if answer != challenge["challenge_answer_expected"]:
            return False
        return True

    def prepare_challenge(self):
        fh_name = None
        try:
            fh = tempfile.NamedTemporaryFile(delete=False)
            self.challenge_file_path = fh.name
            challenge = self.generate_random_string()
            fh.write((challenge + "\r\n").encode("utf-8"))
            fh.close()
            return {
                "challenge_question": self.challenge_file_path,
                "challenge_answer_expected": challenge,
            }
        except:
            pass
        try:
            if fh_name is not None:
                os.remove(fh_name)
        except:
            pass

    def remove_challenge(self):
        if self.challenge_file_path is None:
            return
        try:
            os.remove(self.challenge_file_path)
        except:
            pass
        self.challenge_file_path = None

    def parse_mapping_request(self, mapping_request):
        req_lines = mapping_request.splitlines()
        if len(req_lines) < 2:
            return None
        if req_lines[0].upper().strip() != "MAP/1.0 SET DIR":
            return None
        req_path = req_lines[1].strip()
        if not req_path:
            return None
        if (not os.path.exists(req_path)) or (not os.path.isdir(req_path)):
            return None
        if not os.access(req_path, os.R_OK):
            return None
        return req_path

    def setup(self):
        self.challenge_file_path = None
        self.status_code = None
        self.request_lines = ""
        self.error_messages = []
        self.to_do_logging = self.server.to_do_logging

    def sanitize_message(self, message):
        message = message.replace("\t", "    ")
        return "".join(ch for ch in message if (unicodedata.category(ch)[0] != "C"))

    def finish(self):
        self.remove_challenge()
        if self.to_do_logging:
            request_lines = self.request_lines.splitlines()
            request_start = self.sanitize_message(request_lines[0] if (len(request_lines) > 0) else "")
            request_dir = self.sanitize_message(request_lines[1] if (len(request_lines) > 1) else "")
            try:
                sys.stderr.write("setting - - - [" + time.strftime("%d/%b/%Y %H:%M:%S", time.localtime()) +"] \"" + request_start + "\" " +str(self.status_code) + " -\n")
                sys.stderr.write(request_dir + "\n")
                for message in self.error_messages:
                    sys.stderr.write("setting error: " + message + "\n")
            except:
                pass

    def handle(self):
        self.request.setblocking(True)
        self.request.settimeout(SERVER_SETTING_TIMEOUT)
        try:
            mapping_request = str(self.request.recv(4096), "utf-8")
        except:
            self.error_messages.append("could not read the request line")
            return
        self.request_lines = mapping_request
        asked_path = self.parse_mapping_request(mapping_request)
        if asked_path is None:
            self.status_code = 400
            self.error_messages.append("could not get the requested path")
            try:
                self.request.sendall("MAP/1.0 KO 400\r\n".encode("utf-8"))
            except:
                pass
            return
        challenge = self.prepare_challenge()
        if challenge is None:
            self.status_code = 500
            self.error_messages.append("could not prepare challenge")
            try:
                self.request.sendall("MAP/1.0 KO 500\r\n".encode("utf-8"))
            except:
                pass
            return
        error_occurred = False
        try:
            self.request.sendall(("MAP/1.0 CHALLENGE\r\n" + challenge["challenge_question"] + "\r\n").encode("utf-8"))
            challenge_answer = str(self.request.recv(4096), "utf-8")
        except:
            error_occurred = True
            self.error_messages.append("could not send the challenge or read the answer")
        if error_occurred:
            return
        if not self.check_challenge(challenge, challenge_answer):
            self.status_code = 401
            self.error_messages.append("answer on the challenge was wrong")
            try:
                self.request.sendall("MAP/1.0 KO 401\r\n".encode("utf-8"))
            except:
                pass
        else:
            item = self.update_setting(asked_path)
            if item["is_new"] and (self.server.envelope.wx_frame is not None):
                wx.CallAfter(self.server.envelope.wx_frame.update_session_list)
            self.status_code = 201
            try:
                port_web = str(self.server.envelope.port_web)
                self.request.sendall(
                    ("MAP/1.0 OK 201\r\nhttp://" + SERVER_DOMAIN + ":" + port_web + item["url"] + "\r\n").encode("utf-8")
                )
            except:
                self.error_messages.append("could not send the set url")

class SettingServerEnvelope():
    def __init__(self, port, mapping, session, maplock, logging):
        self.mapping = mapping
        self.session = session
        self.maplock = maplock
        self.logging = logging
        self.wx_frame = None
        self.port = port
        self.port_web = 0
        self.server = None
        self.error_occurred = False

    def prepare(self):
        try:
            self.server = socketserver.ThreadingTCPServer((SERVER_DOMAIN, self.port), SettingRH, False)
            self.server.to_do_logging = self.logging
            self.server.allow_reuse_address = True
            self.server.server_bind()
            self.server.server_activate()
            self.server.envelope = self
        except:
            self.error_occurred = True
        if self.error_occurred:
            if self.server is not None:
                try:
                    self.server.server_close()
                except:
                    pass
                self.server = None
        return not self.error_occurred

    def run(self):
        if self.error_occurred:
            return
        try:
            self.server.serve_forever()
        except:
            self.error_occurred = True
        finally:
            self.server.server_close()

class ServingRHBase(http.server.BaseHTTPRequestHandler):
    extensions_map = http.server.SimpleHTTPRequestHandler.extensions_map
    server_version = "Webmixer"
    sys_version = "v." + str(APPLICATION_VERSION)

    def check_request_allowed(self, requires_referrer):
        host = self.headers.get("Host", None)
        if host is None:
            return False
        if self.server.envelope.localhost_host.match(host) is None:
            return False
        referrer = self.headers.get("Referer", None)
        if requires_referrer and (referrer is None):
            return False
        if referrer is not None:
            if self.server.envelope.localhost_referrer.match(referrer) is None:
                return False
        # Chrome consideres a request to be CORS, thus sending the "Origin" request,
        # even when it is not a cross-site request: JS modules seem to be treated this way;
        # thus requests with this header filled cannot be immediately discarded, sigh;
        origin = self.headers.get("Origin", None)
        if origin is not None:
            if self.server.envelope.localhost_origin.match(origin) is None:
                return False
            # they at least seem to send this info alongside the spurious CORS headers;
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-Site
            if self.headers.get("Sec-Fetch-Site", None) != "same-origin":
                return False
        return True

    def unescape_url_path(self):
        return urllib.parse.unquote_plus(self.path)

    def translate_to_dir(self):
        path = None
        path_to_use = self.unescape_url_path()
        path_to_test = path_to_use
        if not path_to_test.endswith("/"):
            path_to_test += "/"
        envelope = self.server.envelope
        path_start = ""
        path_end = ""
        with envelope.maplock:
            for mapset in [envelope.session, envelope.mapping]:
                for rule in mapset:
                    match = rule["url_re"].match(path_to_test)
                    if match is not None:
                        path_start = rule["dir"]
                        path_end = path_to_use[match.end():]
                        path = path_start + path_end
                        break
                if path is not None:
                    break
        if path is not None:
            if is_hidden(path, trim_subpath(path_end), True, True):
                path = None
        if (path is not None) and path_end:
            if is_dotted_inner(path_end):
                path = None
        return path

    def escape_html_link(self, link):
        return html.escape(link, True).replace(" ", "%20")

    def escape_html_name(self, name):
        return html.escape(name, False)

    def list_directory(self, path):
        dirs = []
        files = []
        if not path.endswith("/"):
            path += "/"
        path_object = pathlib.Path(path)
        base_path = str(path_object)
        base_path_len = len(base_path)
        if not base_path.endswith("/"):
            base_path += "/"
        for item in sorted(path_object.glob("*")):
            rel_path = str(item)[base_path_len:]
            sub_path = trim_subpath(rel_path)
            if not sub_path:
                continue
            if is_hidden(str(item), sub_path, False, item.is_file()):
                continue
            if item.is_dir():
                dirs.append({"rel_path": rel_path, "escaped_link": self.escape_html_link(sub_path) + "/", "escaped_view": self.escape_html_name(sub_path) + "/"})
            if item.is_file():
                if rel_path.split("/")[-1].split(".")[-1] not in ["html", "htm"]:
                    continue
                files.append({"rel_path": rel_path, "escaped_link": self.escape_html_link(sub_path), "escaped_view": self.escape_html_name(sub_path)})
        return {"dirs": dirs, "files": files}

    def write_html_dir_list(self, path):
        self.send_response(200)
        self.send_header("X-Content-Type-Options", "nosniff")
        self.send_header("Referrer-Policy", "same-origin")
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.end_headers()
        dir_content = self.list_directory(path)
        self.wfile.write(
            "<!DOCTYPE html>\n<html><head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n</head>\n<body><pre>\n".encode("utf-8")
        )
        self.wfile.write("Directories:\n".encode("utf-8"))
        self.wfile.write(
            "\n".join([f"<a href=\"{item['escaped_link']}\">{item['escaped_view']}</a>" for item in dir_content["dirs"]]).encode("utf-8")
        )
        self.wfile.write("\nFiles:\n".encode("utf-8"))
        self.wfile.write(
            "\n".join([f"<a href=\"{item['escaped_link']}\">{item['escaped_view']}</a>" for item in dir_content["files"]]).encode("utf-8")
        )
        self.wfile.write(
            "\n</pre></body></html>\n".encode("utf-8")
        )

    def write_file_content(self, path):
        mimetype = http.server.SimpleHTTPRequestHandler.guess_type(self, path)
        if ((mimetype == "application/octet-stream") and (path.endswith("manifest"))):
            mimetype = "text/cache-manifest"
        if (mimetype == "text/html"):
            mimetype += "; charset=utf-8"
        self.send_response(200)
        self.send_header("X-Content-Type-Options", "nosniff")
        self.send_header("Referrer-Policy", "same-origin")
        self.send_header("Content-type", mimetype)
        self.end_headers()
        with open(path, "rb") as file_hnd:
            self.wfile.write(file_hnd.read())

    def do_GET(self):
        has_to_have_referrer = True
        last_path_part = self.path.split("/")[-1].split(".")
        if (len(last_path_part) > 1) and (last_path_part[-1] in ["html", "htm"]):
            has_to_have_referrer = False
        if (len(last_path_part) == 1):
            has_to_have_referrer = False
        if not self.check_request_allowed(has_to_have_referrer):
            try:
                self.send_error(403, "not allowed at all")
            except:
                pass
            return
        path = self.translate_to_dir()
        if not path:
            try:
                self.send_error(404, "respective mapping was not found")
            except:
                pass
            return
        if not os.path.exists(path):
            try:
                self.send_error(404, "mapped path does not exist")
            except:
                pass
            return
        if not os.access(path, os.R_OK):
            try:
                self.send_error(403, "mapped path is not readable")
            except:
                pass
            return
        if os.path.isdir(path):
            try:
                self.write_html_dir_list(path)
            except:
                pass
        else:
            try:
                self.write_file_content(path)
            except:
                pass

class ServingRHQuiet(ServingRHBase):
    def log_message(self, format, *args):
        return
    def log_request(self, code):
        return

ServingRH = {
    True: ServingRHBase,
    False: ServingRHQuiet,
}

class ServingServerEnvelope():
    def __init__(self, port, mapping, session, maplock, logging):
        self.mapping = mapping
        self.session = session
        self.maplock = maplock
        self.logging = logging
        self.wx_frame = None
        self.localhost_referrer = re.compile("http(s)?\\:\\/\\/localhost(\\:[\\d]*)?\\/", re.IGNORECASE)
        self.localhost_origin = re.compile("http(s)?\\:\\/\\/localhost(\\:[\\d]*)?(\\/|$)", re.IGNORECASE)
        self.localhost_host = re.compile("localhost(\\:[\\d]*)?(\\/|$)", re.IGNORECASE)
        self.port = port
        self.server = None
        self.error_occurred = False

    def prepare(self):
        serving_rh_class = ServingRH[self.logging]

        if self.port is not None:
            has_bound = False
            try:
                self.server = http.server.ThreadingHTTPServer((SERVER_DOMAIN, self.port), serving_rh_class)
                self.server.envelope = self
                has_bound = True
            except:
                has_bound = False
            if has_bound:
                self.server.envelope = self
            elif self.server is not None:
                try:
                    self.server.server_close()
                except:
                    pass
                self.server = None
            return has_bound

        while True:
            self.port = random.randint(SERVER_SERVING_PORT_MIN,SERVER_SERVING_PORT_MAX)
            has_bound = False
            try:
                self.server = http.server.ThreadingHTTPServer((SERVER_DOMAIN, self.port), serving_rh_class)
                has_bound = True
            except:
                has_bound = False
            if has_bound:
                self.server.envelope = self
                break
            if self.server is not None:
                try:
                    self.server.server_close()
                except:
                    pass
                self.server = None
        return True

    def run(self):
        try:
            self.server.serve_forever()
        except:
            self.error_occurred = True
        finally:
            self.server.server_close()

def run_setting_server(setting_envelope):
    setting_envelope.run()

def run_serving_server(serving_envelope):
    serving_envelope.run()

def remove_setting_server(setting_envelope):
    if setting_envelope.server is not None:
        try:
            setting_envelope.server.server_close()
        except:
            pass
    del setting_envelope

def prepare_setting_server(port, mapping, session, maplock, logging):
    setting_envelope = SettingServerEnvelope(port, mapping, session, maplock, logging)
    success = setting_envelope.prepare()
    if success:
        return setting_envelope
    else:
        del setting_envelope
        return None

def prepare_serving_server(port, mapping, session, maplock, logging):
    serving_envelope = ServingServerEnvelope(port, mapping, session, maplock, logging)
    success = serving_envelope.prepare()
    if success:
        return serving_envelope
    else:
        del serving_envelope
        return None

class HelpArgFormat(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass

def is_url_session_wise(url):
    return ALLOWED_DYNAMIC_URLS.match(url) is not None

def take_mapping(file_path, is_session):
    mapping = []
    if not file_path:
        return mapping

    first_path = None
    with open(file_path, "r", encoding="utf-8") as file_hnd:
        for line in file_hnd.readlines():
            line = line.strip()
            if ((not line) or (line[0] == "#")):
                continue
            if first_path is None:
                first_path = line
            else:
                if not first_path.endswith("/"):
                    first_path += "/"
                if not line.endswith("/"):
                    line += "/"
                if is_session:
                    if not is_url_session_wise(first_path):
                        first_path = None
                        continue
                mapping.append({
                    "url": first_path,
                    "url_re": re.compile(re.escape(first_path)),
                    "dir": line,
                    "dir_re": re.compile(re.escape(line)),
                })
                first_path = None

    return mapping

def normalize_path(path):
    return os.path.abspath(os.path.expanduser(path)).replace(os.sep, '/')

def take_parameters():
    params = {
        "mapping_path": MAPPING_STATIC_WIN if is_win_form() else MAPPING_STATIC,
        "session_path": MAPPING_SESSION_WIN if is_win_form() else MAPPING_SESSION,
        "port_setting": SERVER_SETTING_PORT,
        "port_serving": SERVER_SERVING_PORT,
        "with_static_map": True,
        "with_dynamic_map": True,
        "with_session": True,
        "logging_web": LOGGING_WEB_DEFAULT,
        "logging_map": LOGGING_SET_DEFAULT,
    }
    help_desc = "\n".join([
        (APPLICATION_NAME + ": local-only web server securely providing multiple directories,"),
        (((len(APPLICATION_NAME) + 2) * " ") + "with both static and dynamic setting of the served directories"),
    ])
    parser = argparse.ArgumentParser(
        description=help_desc,
        epilog="see the " + APPLICATION_NAME + " website:\n" + APPLICATION_LINK,
        formatter_class=HelpArgFormat
    )
    parser.add_argument("--version", help="write version and exit", action="store_true")
    parser.add_argument("--license", help="write license and exit", action="store_true")
    parser.add_argument("--mapping-path", help="file with default directory mapping", default=params["mapping_path"])
    parser.add_argument("--session-path", help="file with saved session mapping", default=params["session_path"])
    parser.add_argument("--port-setting", help="port for the setting actions", type=int, default=SERVER_SETTING_PORT)
    parser.add_argument("--port-serving", help="port for web serving", type=int, default=SERVER_SERVING_PORT)
    parser.add_argument("--port-serving-random", help="use random port for web serving", action="store_true")
    parser.add_argument("--no-static-mapping", help="without static mapping", action="store_true")
    parser.add_argument("--no-dynamic-mapping", help="without dynamic mapping", action="store_true")
    parser.add_argument("--no-session", help="without saving the dynamic mapping", action="store_true")
    parser.add_argument("--restore-session", help="restore the dynamic mapping at startup", action="store_true")
    parser.add_argument("--log-web", help="whether to log web requests", action="store_true", default=LOGGING_WEB_DEFAULT)
    parser.add_argument("--log-set", help="whether to log setting requests", action="store_true", default=LOGGING_SET_DEFAULT)
    args = parser.parse_args()
    to_end = False
    if args.version:
        sys.stderr.write("version: " + APPLICATION_VERSION + "\n")
        to_end = True
    if args.license:
        sys.stderr.write("license: " + APPLICATION_LICENSE + "\n")
        to_end = True
    if to_end:
        sys.exit(0)
    params["logging_web"] = args.log_web
    params["logging_map"] = args.log_set
    params["restore_session"] = args.restore_session
    if args.no_static_mapping:
        params["with_static_map"] = False
        params["mapping_path"] = None
    if args.no_dynamic_mapping:
        params["with_dynamic_map"] = False
        params["port_setting"] = None
        params["session_path"] = None
        params["with_session"] = False
        params["restore_session"] = False
        if args.restore_session:
            sys.stderr.write("session is not restored when dynamic mapping is not used\n")
    if params["with_dynamic_map"] and args.port_setting:
        if ((args.port_setting < PORT_MIN) or (args.port_setting > PORT_MAX)):
            sys.stderr.write(f"port-setting has to be an integer between {PORT_MIN} and {PORT_MAX}\n")
            to_end = True
        else:
            params["port_setting"] = args.port_setting
    if args.port_serving:
        if ((args.port_serving < PORT_MIN) or (args.port_serving > PORT_MAX)):
            sys.stderr.write(f"port-serving has to be an integer between {PORT_MIN} and {PORT_MAX}\n")
            to_end = True
        else:
            params["port_serving"] = args.port_serving
    if args.port_serving_random:
        params["port_serving"] = None
    if (
        params["with_dynamic_map"] and (params["port_serving"] is not None)
        and (params["port_setting"] == params["port_serving"])
    ):
        sys.stderr.write("port-setting and port-serving cannot be the same\n")
        to_end = True
    if to_end:
        sys.exit(2)
    if params["with_dynamic_map"]:
        params["with_session"] = not args.no_session
        if not params["with_session"]:
            params["session_path"] = None
            params["restore_session"] = False
            if args.restore_session:
                sys.stderr.write("session is not restored when session is not used\n")
    if params["with_dynamic_map"] and params["with_session"]:
        norm_session_path = normalize_path(args.session_path)
        session_path_dir = os.path.dirname(norm_session_path)
        try:
            if not os.path.exists(session_path_dir):
                os.makedirs(session_path_dir, mode=0o700, exist_ok=True)
        except:
            pass
        can_write_session = False
        try:
            with open(norm_session_path, "a", encoding="utf-8"):
                can_write_session = True
        except:
            can_write_session = False
        params["session_path"] = norm_session_path if can_write_session else None
        if params["session_path"] is None:
            params["restore_session"] = False
            sys.stderr.write("cannot write to the session file:\n")
            sys.stderr.write(args.session_path + "\n")
            sys.stderr.write("session restore is disabled, set session file by --session-path\n")
            sys.stderr.write("or use '--no-session' option to disable this warning\n")
            sys.stderr.write("or '--no-dynamic-mapping' to disable dynamic mapping at all\n")
    if params["with_static_map"]:
        norm_mapping_path = normalize_path(args.mapping_path)
        if (
            (not os.path.exists(norm_mapping_path))
            or (not os.path.isfile(norm_mapping_path))
            or (not os.access(norm_mapping_path, os.R_OK))
        ):
            sys.stderr.write("cannot read from the file with static mapping:\n")
            sys.stderr.write(args.mapping_path + "\n")
            sys.stderr.write("static mapping '--mapping-path' has to be a readable file,\n")
            sys.stderr.write("use '--no-static-mapping' option to disable static mapping\n")
            to_end = True
        else:
            params["mapping_path"] = norm_mapping_path
    if to_end:
        sys.exit(2)
    return params

def main():
    params = take_parameters()
    session = []
    session_restore = []
    if params["session_path"] is not None:
        last_session = None
        try:
            last_session = take_mapping(params["session_path"], True)
        except:
            last_session = None
            sys.stderr.write(f"cannot read the file with session mapping:\n{params['session_path']}\n")
        if last_session is not None:
            session_restore = last_session
            if params["restore_session"]:
                session = last_session
    mapping = []
    if params["mapping_path"] is not None:
        try:
            mapping = take_mapping(params["mapping_path"], False)
        except:
            sys.stderr.write(f"cannot read the file with static mapping:\n{params['mapping_path']}\n")
            sys.exit(3)
    maplock = threading.Condition()
    setting_envelope = None
    if params["with_dynamic_map"]:
        setting_envelope = prepare_setting_server(params["port_setting"], mapping, session, maplock, params["logging_map"])
        if setting_envelope is None:
            sys.stderr.write("cannot start the setting server: port already in use\n")
            sys.exit(3)
    serving_envelope = prepare_serving_server(params["port_serving"], mapping, session, maplock, params["logging_web"])
    if serving_envelope is None:
        remove_setting_server(setting_envelope)
        sys.stderr.write("cannot start the serving server: port already in use\n")
        sys.exit(3)
    setting_thread = None
    if setting_envelope is not None:
        setting_envelope.port_web = serving_envelope.port
        setting_thread = threading.Thread(target=run_setting_server, args=(setting_envelope,))
        setting_thread.daemon = True
        setting_thread.block_on_close = False
        setting_thread.start()
    serving_thread = threading.Thread(target=run_serving_server, args=(serving_envelope,))
    serving_thread.daemon = True
    serving_thread.block_on_close = False
    serving_thread.start()
    use_xclip = False
    xclip_path = None
    if is_linux_like():
        xclip_path = shutil.which("xclip")
        if (xclip_path):
            use_xclip = True
    app = wx.App()
    properties = {
        "mapping_path": params["mapping_path"],
        "session_path": params["session_path"],
        "mapping": mapping,
        "session": session,
        "session_restore": session_restore,
        "maplock": maplock,
        "with_static_map": params["with_static_map"],
        "with_dynamic_map": params["with_dynamic_map"],
        "setting_envelope": setting_envelope,
        "serving_envelope": serving_envelope,
        "port_setting": params["port_setting"],
        "port_serving": serving_envelope.port,
        "setting_thread": setting_thread,
        "serving_thread": serving_thread,
        "use_xclip": use_xclip,
        "xclip_path": xclip_path,
    }
    frame = AppFrame(properties)
    app.SetTopWindow(frame)
    app.MainLoop()

if __name__ == "__main__":
    main()
